import os
import shutil
import time
from typing import Optional

from airflow import settings as airflow_settings
from airflow.utils.log.logging_mixin import LoggingMixin
from git import Repo

from advai_airflow_plugins.policy.config import PolicyConfigFactory, PolicyConfig
from advai_airflow_plugins.security.manager import TeamRoleManager
from advai_airflow_plugins.validator import Validator


class GitSync(LoggingMixin):

    def __init__(
            self,
            url: str,
            factory: PolicyConfigFactory,
            local_path: str = "/tmp/advai_airflow_dags",
            dags_path: Optional[str] = None,
            interval: int = 60
    ):
        self.url = url
        self.local_path = local_path
        self.dags_path = dags_path or airflow_settings.DAGS_FOLDER
        self.factory = factory
        self.interval = interval
        self._running = False

        if os.path.exists(local_path):
            self.log.info(f"clear local git folder: {local_path}")
            shutil.rmtree(local_path)

        self.log.info(f"clone from {url} to {local_path}")
        self.repo = Repo.clone_from(url, local_path)

        self.log.info(f"git submodule update --init")
        self.repo.git.submodule("update", "--init")

    def _generate_dags_zipfile(self, config: PolicyConfig):
        tmp_file_name = "_tmp"
        tmp_file_name_with_zip = tmp_file_name + ".zip"
        if os.path.exists(tmp_file_name_with_zip):
            self.log.warning(f"remove tmp file {tmp_file_name_with_zip}")
            os.remove(tmp_file_name_with_zip)
        self.log.info(f"make tmp zip archive for role:{config.role_name}, module:{config.git_module}")
        shutil.make_archive(tmp_file_name, "zip", self.local_path + "/" + config.git_module)
        target_dag_path = os.path.join(self.dags_path, config.git_module + ".zip")
        self.log.info(f"move tmp zip archive to dags folder: {target_dag_path}")
        shutil.move(tmp_file_name_with_zip, target_dag_path)

    def _update_git_module(self, config: PolicyConfig) -> bool:
        self.log.info(f"git submodule update --init --remote {config.git_module}")
        self.repo.git.submodule("update", "--init", "--remote", config.git_module)
        diff_list = self.repo.index.diff(None)
        if diff_list is None or (len(diff_list) == 0):
            self.log.info(f"submodule {config.git_module} nothing updated")
            return False

        if len(diff_list) > 1 or diff_list[0].a_path != config.git_module:
            raise Exception("unexpected diff: " + ",".join([str(diff) for diff in diff_list]))

        self.log.info(f"submodule {config.git_module} got updated")
        return True

    def _validate(self) -> bool:
        self.log.info(f"validating dags, path: {self.dags_path}")
        validator = Validator(self.dags_path)
        if validator.has_import_error:
            for file_path, msg in validator.import_errors.items():
                self.log.warning(f"validator error {file_path}: {msg}")
            return False
        self.log.info("validate SUCCESS")
        return True

    def _rollback_git_module(self, config: PolicyConfig):
        self.log.info(f"rollback git module: {config.git_module}")
        self.repo.git.submodule("update", "--init", config.git_module)

    def _commit_git_module(self, config: PolicyConfig):
        self.repo.git.add(config.git_module)
        self.repo.index.commit(f"[GitSync]UPD: module {config.git_module}")
        self.repo.remote().push()
        self.log.info(
            f"Git Commit module: {config.git_module}, commit: {self.repo.submodule(config.git_module).hexsha}")

    def validate_module(self, config: PolicyConfig):
        self.log.info(f"start to validate module: {config.git_module}")
        if not self._update_git_module(config):
            return
        if self._validate():
            self._commit_git_module(config)
        else:
            self._rollback_git_module(config)
            return

    def validate(self):
        for config in self.factory.configs:
            self.validate_module(config)

    def sync_module(self, config: PolicyConfig):
        self.log.info(f"start to fetch and sync module: {config.git_module}")
        if not self._update_git_module(config):
            return
        self._generate_dags_zipfile(config)

    def sync(self):
        for config in self.factory.configs:
            self.sync_module(config)

    def update_perms_module(self, config: PolicyConfig):
        self.log.info(f"start to fetch and update_perms module: {config.git_module}")
        if not self._update_git_module(config):
            return
        TeamRoleManager(config).merge_role_perms()

    def update_perms(self):
        for config in self.factory.configs:
            self.update_perms_module(config)

    def start_validate(self):
        self._running = True
        while self._running:
            self.validate()
            time.sleep(self.interval)

    def start_sync(self):
        self._running = True
        while self._running:
            self.sync()
            time.sleep(self.interval)

    def start_update_perms(self):
        self._running = True
        while self._running:
            self.update_perms()
            time.sleep(self.interval)

    def stop(self, signal_num, frame):
        self.log.warning("Got term signal, stopping!!!")
        self._running = False


if __name__ == '__main__':
    import signal
    import argparse
    from advai_airflow_plugins import config as plugins_config

    parser = argparse.ArgumentParser(description='GitSync command line.')
    parser.add_argument("action", choices=['validate', 'sync', 'update_perms'])
    parser.add_argument('--interval', type=int, default=60,
                        help='GitSync interval in seconds.')
    args = parser.parse_args()
    testConf = PolicyConfig(
        role_name="hh",
        file_name="childhh",
        git_module="child1",
        tag="hi",
        queue="hq",
        rmm_permission="no",
        images=[]
    )
    testFac = PolicyConfigFactory(
        configs=[testConf],
        protected_images=[],
    )
    worker = GitSync(url="https://oauth2:glpat-ejWLzz77CQfoEJpy9Mw8@repo.advai.net/hang.huang/parent.git", factory=testFac, interval=args.interval)
    signal.signal(signal.SIGTERM, worker.stop)
    if args.action == "sync":
        worker.start_sync()
    elif args.action == "validate":
        worker.start_validate()
    elif args.action == "update_perms":
        worker.start_update_perms()
