import os
from typing import List

from airflow import settings
from airflow.models import DagBag
from airflow.security import permissions
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.session import provide_session
from airflow.www.security import ApplessAirflowSecurityManager

from advai_airflow_plugins.policy.config import PolicyConfig


class AirflowPermission:

    def __init__(self, action: str, resource: str):
        self.action = action
        self.resource = resource


class AirflowRolePerms:

    def __init__(self, role: str, perms: List[AirflowPermission]):
        self.role = role
        self.perms = perms


class AdvaiAirflowSecurityManager(ApplessAirflowSecurityManager):

    @provide_session
    def __init__(self, session=None):
        super(AdvaiAirflowSecurityManager, self).__init__(session=session)

    def bulk_merge_role(self, roles: List[AirflowRolePerms]):
        """Sync the provided roles and permissions."""
        existing_roles = self._get_all_roles_with_permissions()
        pvs = self._get_all_non_dag_permissionviews()

        for config in roles:
            role_name = config.role
            perms = config.perms
            role = existing_roles.get(role_name) or self.add_role(role_name)

            new_perm_view_id_set = set()
            for perm_name, view_name in perms:
                perm_view = pvs.get((perm_name, view_name)) or self.add_permission_view_menu(
                    perm_name, view_name
                )
                new_perm_view_id_set.add(perm_view.id)

                if perm_view not in role.permissions:
                    self.log.info(f"Add permission: {perm_view} for role: {role}")
                    self.add_permission_role(role, perm_view)

            delete_permission_role_list = []
            for perm_view in role.permissions:
                if perm_view.id not in new_perm_view_id_set:
                    delete_permission_role_list.append(perm_view)

            for perm_view in delete_permission_role_list:
                self.log.info(f"Delete permission: {perm_view} for role: {role}")
                self.del_permission_role(role, perm_view)


class TeamRoleManager(LoggingMixin):

    basic_perms = [
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_AUDIT_LOG),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG_DEPENDENCIES),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG_CODE),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG_RUN),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_IMPORT_ERROR),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_JOB),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_MY_PASSWORD),
        AirflowPermission(permissions.ACTION_CAN_EDIT, permissions.RESOURCE_MY_PASSWORD),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_MY_PROFILE),
        AirflowPermission(permissions.ACTION_CAN_EDIT, permissions.RESOURCE_MY_PROFILE),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_PLUGIN),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_SLA_MISS),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_TASK_INSTANCE),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_TASK_LOG),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_XCOM),
        AirflowPermission(permissions.ACTION_CAN_READ, permissions.RESOURCE_WEBSITE),
        AirflowPermission(permissions.ACTION_CAN_ACCESS_MENU, permissions.RESOURCE_BROWSE_MENU),
        AirflowPermission(permissions.ACTION_CAN_ACCESS_MENU, permissions.RESOURCE_DAG_RUN),
        AirflowPermission(permissions.ACTION_CAN_ACCESS_MENU, permissions.RESOURCE_DOCS),
        AirflowPermission(permissions.ACTION_CAN_ACCESS_MENU, permissions.RESOURCE_DOCS_MENU),
        AirflowPermission(permissions.ACTION_CAN_ACCESS_MENU, permissions.RESOURCE_JOB),
        AirflowPermission(permissions.ACTION_CAN_ACCESS_MENU, permissions.RESOURCE_AUDIT_LOG),
        AirflowPermission(permissions.ACTION_CAN_ACCESS_MENU, permissions.RESOURCE_PLUGIN),
        AirflowPermission(permissions.ACTION_CAN_ACCESS_MENU, permissions.RESOURCE_SLA_MISS),
        AirflowPermission(permissions.ACTION_CAN_ACCESS_MENU, permissions.RESOURCE_TASK_INSTANCE),
        AirflowPermission(permissions.ACTION_CAN_CREATE, permissions.RESOURCE_TASK_INSTANCE),
        AirflowPermission(permissions.ACTION_CAN_EDIT, permissions.RESOURCE_TASK_INSTANCE),
        AirflowPermission(permissions.ACTION_CAN_DELETE, permissions.RESOURCE_TASK_INSTANCE),
        AirflowPermission(permissions.ACTION_CAN_CREATE, permissions.RESOURCE_DAG_RUN),
        AirflowPermission(permissions.ACTION_CAN_EDIT, permissions.RESOURCE_DAG_RUN),
        AirflowPermission(permissions.ACTION_CAN_DELETE, permissions.RESOURCE_DAG_RUN),
    ]

    def __init__(self, config: PolicyConfig):
        self.config = config
        self.dag_bag = DagBag(
            dag_folder=os.path.join(settings.DAGS_FOLDER, config.file_name),
            include_examples=False,
            include_smart_sensor=False,
            safe_mode=False,
            read_dags_from_db=False,
            store_serialized_dags=False,
            load_op_links=False
        )
        if len(self.dag_bag.import_errors) > 0:
            raise Exception(
                f"Got import errors: \n" + "\n".join([f"{k}: {v}" for k, v in self.dag_bag.import_errors.items()]))

    def generate_dag_perms(self) -> List[AirflowPermission]:
        result = []
        for dag_id in self.dag_bag.dags.keys():
            for action in [permissions.ACTION_CAN_READ, permissions.ACTION_CAN_ACCESS_MENU,
                           permissions.ACTION_CAN_CREATE, permissions.ACTION_CAN_DELETE, permissions.ACTION_CAN_EDIT]:
                result.append(AirflowPermission(action=action, resource=permissions.RESOURCE_DAG_PREFIX + dag_id))
        return result

    def generate_role_perms(self) -> AirflowRolePerms:
        return AirflowRolePerms(self.config.role_name, self.basic_perms + self.generate_dag_perms())

    def merge_role_perms(self):
        manager = AdvaiAirflowSecurityManager()
        manager.bulk_merge_role([self.generate_role_perms()])
